﻿/**
***  キャラクタアニメーションのための人体モデルの表現・基本処理 ライブラリ・サンプルプログラム
***  Copyright (c) 2015-, Masaki OSHITA (www.oshita-lab.org)
***  Released under the MIT license http://opensource.org/licenses/mit-license.php
**/

/**
***  動作接続・遷移アプリケーション
**/

#ifndef  _MOTION_TRANSITION_APP_H_
#define  _MOTION_TRANSITION_APP_H_


// ライブラリ・クラス定義の読み込み
#include "SimpleHuman.h"
#include "SimpleHumanGLUT.h"


// プロトタイプ宣言
class  Timeline;


//
//  動作のメタ情報を表す構造体
//
struct  MotionInfo
{
	// 動作情報
	Motion *  motion;

	// 動作の開始・終了時刻（動作のローカル時間）
	float  begin_time;
	float  end_time;

	// 動作接続・遷移のためのブレンド区間の終了・開始時刻（動作のローカル時間）
	//（ begin_time <= blend_end_time < blend_begin_time <= end_time ）
	float  blend_end_time;
	float  blend_begin_time;

	// キー時刻の配列 [キーフレーム番号]
	vector< float >  keytimes;

	// 次の動作への動作接続のための基準部位
	int  base_segment_no;

	// 描画色
	Color3f  color;
};


//
//  動作遷移の実行のための情報
//
struct  MotionTransitionInfo
{
	// 前後の動作
	const MotionInfo *  prev_motion;
	const MotionInfo *  next_motion;

	// 次の動作の開始時刻（前の動作の開始時刻を基準とするローカル時刻）
	float  next_begin_time;

	// 動作のブレンド開始時刻（前の動作の開始時刻を基準とするローカル時刻）
	float  blend_begin_time;

	// 動作のブレンド終了時刻（前の動作の開始時刻を基準とするローカル時刻）
	float  blend_end_time;

	// 前の動作の変換行列
	Matrix4f  prev_motion_mat;

	// 次の動作の位置・向きを前の動作の位置・向きに合わせるための変換行列
	Matrix4f  next_motion_mat;
};


// 動作のメタ情報を初期化
void  InitMotionInfo( MotionInfo * info, Motion * m = NULL );

// サンプル動作セットの読み込み
const Skeleton *  LoadSampleMotions( vector< MotionInfo * > &  motion_list, const Skeleton * body = NULL );



//
//  動作接続・遷移アプリケーションクラス
//
class  MotionTransitionApp : public GLUTBaseApp
{
  protected:
	// キャラクタ情報

	// キャラクタの骨格
	const Skeleton *   body;

	// キャラクタの姿勢
	Posture *  curr_posture;

  protected:
	// 動作データの情報

	// 動作データリスト
	vector< MotionInfo * >  motion_list;

  protected:
	// 動作再生のための変数

	// 現在の再生動作データ
	Motion *  curr_motion;

	// アニメーション中かどうかを表すフラグ
	bool  on_animation;

	// アニメーションの再生時間
	float  animation_time;

	// アニメーションの再生速度
	float  animation_speed;

	// 現在の表示フレーム番号
	int  frame_no;

	// 現在姿勢の描画色
	Color3f  figure_color;

	// タイムライン描画機能
	Timeline *  timeline;

	// 動作遷移回数のカウント（タイムライン表示用）
	int  transition_count;

  protected:
	// 動作再生（動作遷移・接続）の入力

	// 現在の再生動作番号
	int  curr_motion_no;

	// 次の再生動作番号
	int  next_motion_no;

	// 実行待ちの動作番号
	int  waiting_motion_no;

	// 実行待ちの動作が変更されたかどうかのフラグ
	bool  waiting_motion_changed;

  protected:
	// 動作接続・遷移の設定

	// 動作接続（位置・向き合わせ）を適用するかどうかの設定
	bool  enable_connection;

	// 動作遷移（前後の動作のブレンディング）を適用するかどうかの設定
	bool  enable_transition;

  protected:
	// 動作接続のための変数

	// 現在の再生動作の位置・向きを合わせるための変換行列
	Matrix4f  curr_motion_mat;

	// 現在の動作の再生開始時刻（グローバル時刻）
	float  curr_start_time;

  protected:
	// 動作遷移のための変数

	// 動作遷移の実行のための情報
	MotionTransitionInfo  transition_info;


  public:
	// コンストラクタ
	MotionTransitionApp();

	// デストラクタ
	virtual ~MotionTransitionApp();

  public:
	// イベント処理

	//  初期化
	virtual void  Initialize();

	//  開始・リセット
	virtual void  Start();

	//  画面描画
	virtual void  Display();

	// ウィンドウサイズ変更
	virtual void  Reshape( int w, int h );

	// マウスクリック
	virtual void  MouseClick( int button, int state, int mx, int my );

	// キーボードのキー押下
	virtual void  Keyboard( unsigned char key, int mx, int my );

	// キーボードの特殊キー押下
	virtual void  KeyboardSpecial( unsigned char key, int mx, int my );
	
	// アニメーション処理
	virtual void  Animation( float delta );

  public:
	// 補助処理

	//  次の動作を変更
	void  SetNextMotion( int no = -1 );

	// 動作再生処理（動作接続を考慮、動作遷移は考慮しない）
	void  AnimationWithConnection( float delta );

	// 動作再生処理（動作接続・遷移を考慮）
	void  AnimationWithConnectionTransition( float delta );

	// 動作接続の初期化
	static bool  InitConnection( const MotionInfo * prev_motion, const MotionInfo * next_motion, const Matrix4f & prev_motion_mat, float curr_local_time, 
		MotionTransitionInfo * transition );

	// 動作接続の姿勢計算
	static bool  GetPostureInConnection( const MotionTransitionInfo * transition, float curr_local_time, Posture * posture );

	// 動作接続・遷移の初期化
	static bool  InitTransition( const MotionInfo * prev_motion, const MotionInfo * next_motion, const Matrix4f & prev_motion_mat, float curr_local_time, 
		MotionTransitionInfo * transition );

	// 動作接続・遷移の姿勢計算
	static bool  GetPostureInTransition( const MotionTransitionInfo * transition, float curr_local_time, 
		Posture * posture, float * output_blend_ratio = NULL );

	// タイムラインへの前後の動作・遷移区間の設定
	void  InitTimeline( Timeline & timeline, float prev_start_time, const MotionTransitionInfo & trans );

	// タイムラインへの現在時刻・時間範囲の設定
	void  UpdateTimeline( Timeline & timeline, float curr_time );
};


// 補助処理（グローバル関数）のプロトタイプ宣言

// 変換行列の水平向き（方位角）成分を計算（Ｚ軸の正の方向を０とする時計回りの角度を -180～180 の範囲で求める）
float  ComputeOrientation( const Matrix3f & ori );

// ２つの姿勢の位置・向きを合わせるための変換行列を計算
//（next_frame の位置・向きを、prev_frame の位置・向きに合わせるための変換行列 trans_mat を計算）
void  ComputeConnectionTransformation( const Matrix4f & prev_frame, const Matrix4f & next_frame, Matrix4f & trans_mat );

// ２つの姿勢の位置・向きを合わせるための変換行列を計算
// ２つの姿勢の腰の水平位置・水平向きを合わせる
//（next_pose の腰の水平位置・水平向きを、prev_pose の腰の水平位置・水平向きに合わせる）
void  ComputeConnectionTransformation( const Posture & prev_pose, const Posture & next_pose, Matrix4f & trans_mat );

// ２つの姿勢の位置・向きを合わせるための変換行列を計算
// ２つの姿勢の基準部位の水平位置と、入力として与えられた水平向きを合わせる
//（next_pose の基準部位 base_segment の水平位置を prev_pose の水平位置に合わせて、水平向き next_ori を 水平向き prev_ori に合わせる）
void  ComputeConnectionTransformation( const Posture & prev_pose, float prev_ori, const Posture & next_pose, float next_ori, int base_segment, Matrix4f & trans_mat );

// 姿勢の位置・向きに変換行列を適用
void  TransformPosture( const Matrix4f & trans, Posture & posture );

// 姿勢補間（２つの姿勢を補間）（※レポート課題）
void  MyPostureInterpolation( const Posture & p0, const Posture & p1, float ratio, Posture & p );


#endif // _MOTION_TRANSITION_APP_H_
